package weka.clusterers.FOptics;

import java.io.Serializable;
import java.util.ArrayList;

import weka.clusterers.de.lmu.ifi.dbs.elki.utilities.datastructures.heap.UpdatableHeap;
import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.array.*;

/* 
Copyright (C) 2014
Johannes Schneider, ABB Research,Switzerland, johannes.schneider@alumni.ethz.ch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * FOptics algorithm (Fast approximation of OPTICS)
 * 
 * @author Johannes Schneider
 * 
 */
public class FOptics implements Serializable{

	static final long serialVersionUID = 274552680222105223L;
	
	
	/**
     * undefined value for (reachability/average) distance
     */
    public static final float undefinedDist = -0.1f;
	
    
    /**
     * Result: output order of points
     */
    private int[] order;
    
    /**
     * Result: reachability distances
     */
	private float reachDist[];
    
	
	/**
     * processed points
     */
	boolean processed[];
	
	/**
     * index of next point in order array
     */
	int iorder;
	
	/**
     * neighbors of a point
     */
	ArrayList<TIntList> neighs; 
	
	
	/**
     * Inverse Densities correspond to average distances in point set of projections 
     */
	float [] inverseDensities;

	/**
     * Point set 
     */
	float [][] points;



/**
* Return reachability distances of points
* undefined Distances are set to (negative) constant undefinedDist
* 
* @return reachability distances
*/
public float[] getReachabilityDistances() {
		return reachDist;
	}

/**
* Return order of points
* 
* @return order of points
*/
public int[] getPointOrder() {
		return order;
	}

/**
* Run the algorithm on a matrix of points "points[i]" gives coordinates of point i
* 
* @param points to run the algorithm on
* @param minPts -- similar to OPTICS minPts
*/
public void run(float[][] points,int minPts)
{
	if (points==null || points.length==0)
		return;

	//initialize points used and reachability distance
	this.points=points;	
	int N = points.length;//number of points
    TIntList ptInd = new TIntArrayList(N);
	for (int k=0;k<N;k++)
		ptInd.add(k);
	reachDist = new float[N];
	for (int i=0;i<N;i++)
		reachDist[i]=undefinedDist;

	 //compute projections, density estimates and neighborhoods
	RandProNeighsAndDensities algo = new RandProNeighsAndDensities(points,minPts);
	algo.computeSetsBounds(ptInd); 	//project points	
	inverseDensities = algo.computeAverageDistInSet();	//compute densities
	neighs = algo.getNeighs(); //get neighbors of points	
	
	//compute ordering as for OPTICS 
	iorder=0;
	processed  = new boolean[N];
	order = new int[N];
	for(int ipt=0;ipt<N;ipt++) {
        if(!processed[ipt]) {
          expandClusterOrder(ipt);
        }
      }

}

/**
* OPTICS Algorithm for processing a point -- implemented as for OPTICS (see code in ELKI), but with different density estimates
* 
* @param index of point to process
*/
protected void expandClusterOrder(int ipt) {
    UpdatableHeap<ClusterOrderEntry> heap = new UpdatableHeap<ClusterOrderEntry>();
    ClusterOrderEntry tmp = new GenericClusterOrderEntry(ipt, 0, 1e6f);
    heap.add(tmp);
    while(!heap.isEmpty()) {
      final ClusterOrderEntry current = heap.poll();
      int currPt= current.getID();
      order[iorder]=currPt;
      iorder++;
      processed[currPt]=true;	
      float coredist=inverseDensities[currPt];
      TIntIterator it = neighs.get(currPt).iterator();
  	  while (it.hasNext())
  	  {
  		int iNeigh =it.next();
		if (processed[iNeigh])
			continue;
		float nrdist = VectorOps.getDist(points[iNeigh],points[currPt]);		
		if (coredist>nrdist)		
			nrdist=coredist;  	 
		if (reachDist[iNeigh]==undefinedDist)
			reachDist[iNeigh]=nrdist;
		else if (nrdist< reachDist[iNeigh])
			reachDist[iNeigh]=nrdist;
		tmp= new GenericClusterOrderEntry(iNeigh, currPt,nrdist);
        heap.add(tmp);
  	  }
    }
  }


}
