package weka.clusterers.FOptics;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;

/* 
Copyright (C) 2014
Johannes Schneider, ABB Research,Switzerland, johannes.schneider@alumni.ethz.ch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/**
 * Random Projections used for computing neighbors and density estimates
 * 
 * @author Johannes Schneider
 * 
 */
public class RandProNeighsAndDensities implements Serializable {

	//constants
	private static final long serialVersionUID = 1L;	
	
	/**
     * Default Seed for random numbers
     */
	private static final int DefaultRandSeed = 4711;	//seed for random projections
	
	/**
     * Default constant  used to compute number of projections as well as number of splits of point set, ie. constant *log N*d
     */
	private static final int logOProjectionConst = 20;	//constant in O(log N*d) used to compute number of projections as well as number of splits of point set
	
	/**
     * Sets used for neighborhood computation should be about minSplitSize
     * Sets are still used if they deviate by less (1+/- sizeTolerance)
     */
	private static final float sizeTolerance = 2f/3;	
	
	
	//variables
	/**
     * minimum size for which a point set is further partitioned (roughly corresponds to minPts in OPTICS)
     */
	int minSplitSize;
	
	/**
     * entire point set
     */
	float[][] points; 
	
	/**
     * random number generator
     */	
	Random rand;
	
	/**
     * number of times the point set is projected onto a 1d random line, typically something like O(log N)
     */
	protected int nProject1d;
	
	/**
     * number of times the entire point set is split, typically something like O(log N)
     */
	protected int nPointSetSplits; 
	
	/**
     * sets that resulted from recursive split of entire point set
     */
	protected ArrayList<int[]> splitsets;
	
	/**
     * all projected points
     */
	protected float[][] projectedPoints;
	
	/**
     * number of points
     */
	int N; 
	
	/**
     * number of dimensions
     */
	int nDimensions;
	


	
/**
* Constructor
* 
* @param points to process
* @param minimum size for which a point set is further partitioned (roughly corresponds to minPts in OPTICS)
*/
public RandProNeighsAndDensities(float[][] points, int minSplitSize) {
	this.minSplitSize=minSplitSize;
	if (points==null)
	{
		N=0;
		nDimensions=0;
		this.nProject1d = 0;
		return;
	}
	else
	{
		N=points.length; 
		nDimensions = points[0].length;
		this.points=points;//new float[N][nDimensions];
		
		nPointSetSplits = (int) (logOProjectionConst*Math.log(N*nDimensions+1)/Math.log(2)); //perform O(log N+log dim) splits of the entire point sets projections
		nProject1d= (int) (logOProjectionConst*Math.log(N*nDimensions+1)/Math.log(2)); //perform O(log N+log dim) projections of the point set onto a random line
	}
	rand = new Random(DefaultRandSeed);
	//System.out.println("NPts:"+N + "  dim:"+di);
}

/**
* Random Seed Setting (optional to use)
* 
* @param random seed 
*/
public void setRandomSeed(long seed)
{
	rand = new Random(seed);
}



/**
* Set number of times entire point set is split(optional to use)
* 
* @param number of Splits 
*/
public void setNumberPointSetSplits(int nSplits) 
{
	this.nPointSetSplits = nSplits;
}




/**
* Create random projections, project points and put points into sets of size about minSplitSize/2
* 
* @param indexes of points in point set that are projected, typically 0..N-1
*/
public void computeSetsBounds(TIntList ptList)
{
	    splitsets = new ArrayList<int[]>();
	    
		//perform projections of points
	    projectedPoints = new float[nProject1d][N];	
	    float[][] tmpPro = new float[nProject1d][];
	
		for(int j=0;j<nProject1d;j++)
		    {
		    	float sum=0;
		    	float[] currRp= new float[nDimensions];
		    	for(int i=0;i<nDimensions;i++)		    
			    {
		    		float fl =rand.nextFloat()-0.5f;
			    	currRp[i]=fl;
			    	sum+=fl*fl;		    	
			    }
		    	sum=(float) Math.sqrt(sum);
		    	for(int i=0;i<nDimensions;i++)		    
			    {	    	
		    		currRp[i]/=sum;		    		    	
			    }		    
		    	float[] currPro=projectedPoints[j];
		    	

		    	TIntIterator it = ptList.iterator();
		    	int k=0;
		    	while (it.hasNext())			    
			    {		    	
			    	float[] vecPt = points[it.next()];
			    	sum=0;
			    	for(int i=0;i<nDimensions;i++)
			    	{
			    		sum+=currRp[i]*vecPt[i];	    		
			    	}
			    	currPro[k]=sum;
			    	k++;
			    }
		    }	 	  
		
		//split entire point set, reuse projections by shuffling them 
		TIntArrayList proind = new TIntArrayList(nProject1d);
		for (int j=0;j<nProject1d;j++)
			proind.add(j);    	
	    for (int avgP=0;avgP<nPointSetSplits;avgP++)
	    {  		
	    	//shuffle projections
	    	for(int i=0;i<nProject1d;i++)
	    		tmpPro[i]=projectedPoints[i];	    
	    	proind.shuffle(rand);
	    	TIntIterator it = proind.iterator();	    	 
	    	int i=0;
	    	 while (it.hasNext())
	    	 {
	    		 int cind=it.next();
	    		 projectedPoints[cind]=tmpPro[i];
	    		 i++;
	    	 }	    	
	    	 
		    //split point set
	       int npts=ptList.size();	
		   int[] ind = new int[npts]; 	
		   for(int l=0;l<npts;l++)
		    	ind[l]=l;
		   splitupNoSort(ind,0);
	    }
}

/**
* Recursively splits entire point set until the set is below a threshold

* 
* @param indexes of points in point set that are in the current set
* @param depth of projection (how many times point set has been split already)
*/
public void splitupNoSort(int[] ind,int dim) {
	  int nele=ind.length;	  
	  dim=dim%nProject1d;//choose a projection of points 
	  float[] tpro =projectedPoints[dim];
	  int splitpos;  //position used for splitting the projected points into two sets used for recursive splitting
	  
	  //save set such that used for density or neighborhood computation
	  if (nele > minSplitSize*(1-sizeTolerance)  && nele < minSplitSize*(1+sizeTolerance)) //sets should be roughly minSplitSize 
	  {
					 float[] cpro = new float[nele];
					 for(int i=0;i<nele;i++)
						  cpro[i]=tpro[ind[i]];
					  VectorOps.quicksort(cpro,ind, 0,nele-1); //sort set, since need median element later
					  splitsets.add(ind);
	  }
		 
	  //compute splitting element
	  if (nele > minSplitSize ) // do not store set or even sort set, since it is too large	
		  {
		  	//splits can be performed either by distance (between min,maxCoord) or by picking a point randomly(picking index of point)
		    //outcome is similar
		  
		      //pick random splitting point based on distance
		  	 /* float rmin=Float.MAX_VALUE/2;
		  	  float rmax=-Float.MAX_VALUE/2;
			  int minInd=0;
			  int maxInd=nele-1;		  
			  for(int i=0;i<nele;i++)
			  {
				  float currEle=tpro[ind[i]];
				  rmin=Math.min(currEle, rmin);
				  rmax=Math.max(currEle, rmax);		
			  }
			  
			  if (rmin!=rmax) //if not all elements are the same  
			  {
				  float rs= rmin+rand.nextFloat()*(rmax-rmin);		
				
				  //permute elements such that all points smaller than the splitting element are on the right and the others on the left in the array
				  while (minInd<maxInd)
				  {		  	  
					  float currEle=tpro[ind[minInd]];		  
					  if (currEle>rs)					  
					  {
						  while (minInd<maxInd && tpro[ind[maxInd]]>rs)
							  maxInd--;
						   if (minInd==maxInd)
							   break;
						   int currInd=ind[minInd];
						   ind[minInd]=ind[maxInd];
						   ind[maxInd]=currInd;
						   maxInd--;			  
					  }
					  minInd++;
				  }
				  
			  }
			  else //if all elements are the same split in the middle
				  minInd= nele/2;*/
			
			  
			  //pick random splitting element based on position
			  int randInd =rand.nextInt(nele); //element less than nele
			  float rs= tpro[ind[randInd]];
			  int minInd=0;
			  int maxInd=nele-1;
			  //permute elements such that all points smaller than the splitting element are on the right and the others on the left in the array
			  while (minInd<maxInd)
			  {		  	  
				  float currEle=tpro[ind[minInd]];		  
				  if (currEle>rs)					  
				  {
					  while (minInd<maxInd && tpro[ind[maxInd]]>rs)
						  maxInd--;
					   if (minInd==maxInd)
						   break;
					   int currInd=ind[minInd];
					   ind[minInd]=ind[maxInd];
					   ind[maxInd]=currInd;
					   maxInd--;			  
				  }
				  minInd++;
			  }			  
			  if (minInd==nele-1)
				  minInd=nele/2;
			  
			  
			  //split set recursively
			  splitpos=minInd+1;
		      int[] ind2=new int[splitpos];
		      for(int l=0;l<splitpos;l++)
		    	  ind2[l]=ind[l];
		      splitupNoSort(ind2,dim+1);
		      
		      ind2=new int[nele-splitpos];
		      for(int l=0;l<nele-splitpos;l++)
		    	  ind2[l]=ind[l+splitpos];
		      splitupNoSort(ind2,dim+1);

		  }
	
}


/**
* Compute list of neighbors for each point from sets resulting from projection 
* 
* @return list of neighbors for each point
*/
public ArrayList<TIntList> getNeighs()
{
	//init lists
	ArrayList<TIntList> neighs = new ArrayList<TIntList>(N);
	for(int l=0;l<N;l++)
	{	    	
		TIntList list= new TIntArrayList();	    	
		neighs.add(list);
	}
	
	//go through all sets
	Iterator<int[]> it1 = splitsets.iterator();
	while (it1.hasNext())
	{		
		int[] pinSet = it1.next();
		
		//for each set, ie. each projection line
		int len = pinSet.length;
		int ind=pinSet[0];		
		int indoff=Math.round(len/2); //middle point of projection
		int oldind=pinSet[indoff];
		
		//add all points as neighbors to middle point and the the middle point to all other points in set
		for (int i=0;i<len;i++)
		{				
			ind = pinSet[i];
			TIntList cneighs=neighs.get(ind);
	    	int pos = cneighs.binarySearch(oldind);	//only add point if not a neighbor already
			if (pos<0) // element not found
			{
				if (-pos>cneighs.size()) 
									cneighs.add(oldind); //append at end
				else				
									cneighs.insert(-pos-1,oldind);							
			}				
			TIntList cneighs2=neighs.get(oldind);					
			pos = cneighs2.binarySearch(ind);
		  	if (pos<0) // element not found
			{					
				if (-pos>cneighs2.size())
									cneighs2.add(ind); //append at end
				else				
									cneighs2.insert(-pos-1,ind);						
			}						
		}
	}
	return neighs;
}



/**
* Compute for each point a density estimate as inverse of average distance to a point in a projected set 
* 
* @return for each point average distance to point in a set
*/
public float[] computeAverageDistInSet() 
{
		float [] davg = new float[N];
		int nDists[] = new int[N];		
		Iterator<int[]> it1 = splitsets.iterator();
		while (it1.hasNext()) 
		{									
			int[] pinSet = it1.next();
			int len = pinSet.length;   	
			int indoff=Math.round(len/2);
			int oldind=pinSet[indoff];
			for (int i=0;i<len;i++)
			{	    				
				int ind=pinSet[i];
				if (ind==indoff)
					continue;
				float dist = VectorOps.getDist(points[ind],points[oldind]);
				davg[oldind]+=dist;
				nDists[oldind]++;
				davg[ind]+=dist;
				nDists[ind]++;
			}	    		
		}
		for(int l=0;l<N;l++)
		{	    			
			if (nDists[l]==0)  //it might be that a point does not occur for a certain size of a projection (likely if do few projections, in this case there is no avg distance)
				davg[l]=FOptics.undefinedDist;
			else
				davg[l]/=nDists[l];
		} 
	return davg;
}

}



